@@ -11,25 +11,24 @@ const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 const St = imports.gi.St;
 
-const AppIcon = imports.ui.appIcon;
 const Main = imports.ui.main;
+const Tweener = imports.ui.tweener;
 
 const POPUP_ARROW_COLOR = new Clutter.Color();
 POPUP_ARROW_COLOR.from_pixel(0xffffffff);
+const POPUP_UNFOCUSED_ARROW_COLOR = new Clutter.Color();
+POPUP_UNFOCUSED_ARROW_COLOR.from_pixel(0x808080ff);
 const TRANSPARENT_COLOR = new Clutter.Color();
 TRANSPARENT_COLOR.from_pixel(0x00000000);
-const POPUP_SEPARATOR_COLOR = new Clutter.Color();
-POPUP_SEPARATOR_COLOR.from_pixel(0x80808066);
 
 const POPUP_APPICON_SIZE = 96;
 const POPUP_LIST_SPACING = 8;
 
-const POPUP_POINTER_SELECTION_THRESHOLD = 3;
+const DISABLE_HOVER_TIMEOUT = 500; // milliseconds
 
 const THUMBNAIL_SIZE = 256;
-const THUMBNAIL_POPUP_TIME = 1000; // milliseconds
-
-const HOVER_TIME = 500; // milliseconds
+const THUMBNAIL_POPUP_TIME = 500; // milliseconds
+const THUMBNAIL_FADE_TIME = 0.2; // seconds
 
 function mod(a, b) {
     return (a + b) % b;
@@ -52,15 +51,20 @@ AltTabPopup.prototype = {
         this._haveModal = false;
 
         this._currentApp = 0;
-        this._currentWindows = [];
+        this._currentWindow = 0;
         this._thumbnailTimeoutId = 0;
+        this._motionTimeoutId = 0;
+
+        // Initially disable hover so we ignore the enter-event if
+        // the switcher appears underneath the current pointer location
+        this._disableHover();
 
         global.stage.add_actor(this.actor);
     },
 
     show : function(backward) {
-        let appMonitor = Shell.AppMonitor.get_default();
-        let apps = appMonitor.get_running_apps ("");
+        let tracker = Shell.WindowTracker.get_default();
+        let apps = tracker.get_running_apps ("");
 
         if (!apps.length)
             return false;
@@ -72,44 +76,40 @@ AltTabPopup.prototype = {
         this._keyPressEventId = global.stage.connect('key-press-event', Lang.bind(this, this._keyPressEvent));
         this._keyReleaseEventId = global.stage.connect('key-release-event', Lang.bind(this, this._keyReleaseEvent));
 
-        this._motionEventId = this.actor.connect('motion-event', Lang.bind(this, this._mouseMoved));
-        this._mouseActive = false;
-        this._mouseMovement = 0;
+        this.actor.connect('button-press-event', Lang.bind(this, this._clickedOutside));
+        this.actor.connect('scroll-event', Lang.bind(this, this._onScroll));
 
         this._appSwitcher = new AppSwitcher(apps);
         this.actor.add_actor(this._appSwitcher.actor);
         this._appSwitcher.connect('item-activated', Lang.bind(this, this._appActivated));
-        this._appSwitcher.connect('item-hovered', Lang.bind(this, this._appHovered));
+        this._appSwitcher.connect('item-entered', Lang.bind(this, this._appEntered));
 
-        let primary = global.get_primary_monitor();
-        this._appSwitcher.actor.x = primary.x + Math.floor((primary.width - this._appSwitcher.actor.width) / 2);
-        this._appSwitcher.actor.y = primary.y + Math.floor((primary.height - this._appSwitcher.actor.height) / 2);
+        let focus = global.get_focus_monitor();
+        this._appSwitcher.actor.x = focus.x + Math.floor((focus.width - this._appSwitcher.actor.width) / 2);
+        this._appSwitcher.actor.y = focus.y + Math.floor((focus.height - this._appSwitcher.actor.height) / 2);
 
         this._appIcons = this._appSwitcher.icons;
 
-        // _currentWindows give the index of the selected window for
-        // each app; they all start at 0.
-        this._currentWindows = this._appIcons.map(function (app) { return 0; });
-
         // Make the initial selection
         if (this._appIcons.length == 1) {
-            if (!backward && this._appIcons[0].windows.length > 1) {
+            if (!backward && this._appIcons[0].cachedWindows.length > 1) {
                 // For compatibility with the multi-app case below
-                this._select(0, 1);
+                this._select(0, 1, true);
             } else
                 this._select(0);
         } else if (backward) {
             this._select(this._appIcons.length - 1);
         } else {
-            if (this._appIcons[0].windows.length > 1) {
-                let curAppNextWindow = this._appIcons[0].windows[1];
-                let nextAppWindow = this._appIcons[1].windows[0];
+            let firstWindows = this._appIcons[0].cachedWindows;
+            if (firstWindows.length > 1) {
+                let curAppNextWindow = firstWindows[1];
+                let nextAppWindow = this._appIcons[1].cachedWindows[0];
 
                 // If the next window of the current app is more-recently-used
                 // than the first window of the next app, then select it.
                 if (curAppNextWindow.get_workspace() == global.screen.get_active_workspace() &&
                     curAppNextWindow.get_user_time() > nextAppWindow.get_user_time())
-                    this._select(0, 1);
+                    this._select(0, 1, true);
                 else
                     this._select(1);
             } else
@@ -138,41 +138,50 @@ AltTabPopup.prototype = {
     },
 
     _nextWindow : function() {
-        return mod(this._currentWindows[this._currentApp] + 1,
-                   this._appIcons[this._currentApp].windows.length);
+        return mod(this._currentWindow + 1,
+                   this._appIcons[this._currentApp].cachedWindows.length);
     },
     _previousWindow : function() {
-        return mod(this._currentWindows[this._currentApp] - 1,
-                   this._appIcons[this._currentApp].windows.length);
+        return mod(this._currentWindow - 1,
+                   this._appIcons[this._currentApp].cachedWindows.length);
     },
 
     _keyPressEvent : function(actor, event) {
         let keysym = event.get_key_symbol();
         let shift = (Shell.get_event_state(event) & Clutter.ModifierType.SHIFT_MASK);
 
+        this._disableHover();
+
         // The WASD stuff is for debugging in Xephyr, where the arrow
         // keys aren't mapped correctly
 
-        if (keysym == Clutter.Tab)
-            this._select(shift ? this._previousApp() : this._nextApp());
-        else if (keysym == Clutter.grave)
+        if (keysym == Clutter.grave)
             this._select(this._currentApp, shift ? this._previousWindow() : this._nextWindow());
         else if (keysym == Clutter.Escape)
             this.destroy();
-        else if (this._thumbnails) {
-            if (keysym == Clutter.Left || keysym == Clutter.a)
+        else if (this._thumbnailsFocused) {
+            if (keysym == Clutter.Tab) {
+                if (shift && this._currentWindow == 0)
+                    this._select(this._previousApp());
+                else if (!shift && this._currentWindow == this._appIcons[this._currentApp].cachedWindows.length - 1)
+                    this._select(this._nextApp());
+                else
+                    this._select(this._currentApp, shift ? this._previousWindow() : this._nextWindow());
+            } else if (keysym == Clutter.Left || keysym == Clutter.a)
                 this._select(this._currentApp, this._previousWindow());
             else if (keysym == Clutter.Right || keysym == Clutter.d)
                 this._select(this._currentApp, this._nextWindow());
             else if (keysym == Clutter.Up || keysym == Clutter.w)
                 this._select(this._currentApp, null, true);
         } else {
-            if (keysym == Clutter.Left || keysym == Clutter.a)
+            if (keysym == Clutter.Tab)
+                this._select(shift ? this._previousApp() : this._nextApp());
+            else if (keysym == Clutter.Left || keysym == Clutter.a)
                 this._select(this._previousApp());
             else if (keysym == Clutter.Right || keysym == Clutter.d)
                 this._select(this._nextApp());
             else if (keysym == Clutter.Down || keysym == Clutter.s)
-                this._select(this._currentApp, this._currentWindows[this._currentApp]);
+                this._select(this._currentApp, this._currentWindow);
         }
 
         return true;
@@ -187,45 +196,87 @@ AltTabPopup.prototype = {
         return true;
     },
 
-    _appActivated : function(appSwitcher, n) {
-        Main.activateWindow(this._appIcons[n].windows[this._currentWindows[n]]);
+    _onScroll : function(actor, event) {
+        let direction = event.get_scroll_direction();
+        if (direction == Clutter.ScrollDirection.UP) {
+            if (this._thumbnailsFocused) {
+                if (this._currentWindow == 0)
+                    this._select(this._previousApp());
+                else
+                    this._select(this._currentApp, this._previousWindow());
+            } else {
+                let nwindows = this._appIcons[this._currentApp].cachedWindows.length;
+                if (nwindows > 1)
+                    this._select(this._currentApp, nwindows - 1);
+                else
+                    this._select(this._previousApp());
+            }
+        } else if (direction == Clutter.ScrollDirection.DOWN) {
+            if (this._thumbnailsFocused) {
+                if (this._currentWindow == this._appIcons[this._currentApp].cachedWindows.length - 1)
+                    this._select(this._nextApp());
+                else
+                    this._select(this._currentApp, this._nextWindow());
+            } else {
+                let nwindows = this._appIcons[this._currentApp].cachedWindows.length;
+                if (nwindows > 1)
+                    this._select(this._currentApp, 0);
+                else
+                    this._select(this._nextApp());
+            }
+        }
+    },
+
+    _clickedOutside : function(actor, event) {
         this.destroy();
     },
 
-    _appHovered : function(appSwitcher, n) {
+    _appActivated : function(appSwitcher, n) {
+        // If the user clicks on the selected app, activate the
+        // selected window; otherwise (eg, they click on an app while
+        // !mouseActive) activate the first window of the clicked-on
+        // app.
+        let window = (n == this._currentApp) ? this._currentWindow : 0;
+        Main.activateWindow(this._appIcons[n].cachedWindows[window]);
+        this.destroy();
+    },
+
+    _appEntered : function(appSwitcher, n) {
         if (!this._mouseActive)
             return;
 
-        this._select(n, this._currentWindows[n]);
+        this._select(n);
     },
 
     _windowActivated : function(thumbnailList, n) {
-        Main.activateWindow(this._appIcons[this._currentApp].windows[n]);
+        Main.activateWindow(this._appIcons[this._currentApp].cachedWindows[n]);
         this.destroy();
     },
 
-    _windowHovered : function(thumbnailList, n) {
+    _windowEntered : function(thumbnailList, n) {
         if (!this._mouseActive)
             return;
 
         this._select(this._currentApp, n);
     },
 
-    _mouseMoved : function(actor, event) {
-        if (++this._mouseMovement < POPUP_POINTER_SELECTION_THRESHOLD)
-            return;
+    _disableHover : function() {
+        this._mouseActive = false;
 
-        this.actor.disconnect(this._motionEventId);
+        if (this._motionTimeoutId != 0)
+            Mainloop.source_remove(this._motionTimeoutId);
+
+        this._motionTimeoutId = Mainloop.timeout_add(DISABLE_HOVER_TIMEOUT, Lang.bind(this, this._mouseTimedOut));
+    },
+
+    _mouseTimedOut : function() {
+        this._motionTimeoutId = 0;
         this._mouseActive = true;
-
-        this._appSwitcher.checkHover();
-        if (this._thumbnails)
-            this._thumbnails.checkHover();
     },
 
     _finish : function() {
         let app = this._appIcons[this._currentApp];
-        let window = app.windows[this._currentWindows[this._currentApp]];
+        let window = app.cachedWindows[this._currentWindow];
         Main.activateWindow(window);
         this.destroy();
     },
@@ -243,15 +294,41 @@ AltTabPopup.prototype = {
         if (this._keyReleaseEventId)
             global.stage.disconnect(this._keyReleaseEventId);
 
+        if (this._motionTimeoutId != 0)
+            Mainloop.source_remove(this._motionTimeoutId);
         if (this._thumbnailTimeoutId != 0)
             Mainloop.source_remove(this._thumbnailTimeoutId);
     },
 
-    _select : function(app, window, noTimeout) {
-        if ((app != this._currentApp || !window) && this._thumbnails) {
-            this._thumbnails.actor.destroy();
-            this._thumbnails = null;
-            this._appSwitcher.showArrow(-1);
+    /**
+     * _select:
+     * @app: index of the app to select
+     * @window: (optional) index of which of @app's windows to select
+     * @forceAppFocus: optional flag, see below
+     *
+     * Selects the indicated @app, and optional @window, and sets
+     * this._thumbnailsFocused appropriately to indicate whether the
+     * arrow keys should act on the app list or the thumbnail list.
+     *
+     * If @app is specified and @window is unspecified or %null, then
+     * the app is highlighted (ie, given a light background), and the
+     * current thumbnail list, if any, is destroyed. If @app has
+     * multiple windows, and @forceAppFocus is not %true, then a
+     * timeout is started to open a thumbnail list.
+     *
+     * If @app and @window are specified (and @forceAppFocus is not),
+     * then @app will be outlined, a thumbnail list will be created
+     * and focused (if it hasn't been already), and the @window'th
+     * window in it will be highlighted.
+     *
+     * If @app and @window are specified and @forceAppFocus is %true,
+     * then @app will be highlighted, and @window outlined, and the
+     * app list will have the keyboard focus.
+     */
+    _select : function(app, window, forceAppFocus) {
+        if (app != this._currentApp || window == null) {
+            if (this._thumbnails)
+                this._destroyThumbnails();
         }
 
         if (this._thumbnailTimeoutId != 0) {
@@ -259,37 +336,42 @@ AltTabPopup.prototype = {
             this._thumbnailTimeoutId = 0;
         }
 
+        this._thumbnailsFocused = (window != null) && !forceAppFocus;
+
         this._currentApp = app;
-        if (window != null) {
-            this._appSwitcher.highlight(-1);
-            this._appSwitcher.showArrow(app);
-        } else {
-            this._appSwitcher.highlight(app);
-            if (this._appIcons[this._currentApp].windows.length > 1)
-                this._appSwitcher.showArrow(app);
-        }
+        this._currentWindow = window ? window : 0;
+        this._appSwitcher.highlight(app, this._thumbnailsFocused);
 
         if (window != null) {
             if (!this._thumbnails)
                 this._createThumbnails();
-            this._currentWindows[this._currentApp] = window;
-            this._thumbnails.highlight(window);
-        } else if (this._appIcons[this._currentApp].windows.length > 1 &&
-                   !noTimeout) {
+            this._currentWindow = window;
+            this._thumbnails.highlight(window, forceAppFocus);
+        } else if (this._appIcons[this._currentApp].cachedWindows.length > 1 &&
+                   !forceAppFocus) {
             this._thumbnailTimeoutId = Mainloop.timeout_add (
                 THUMBNAIL_POPUP_TIME,
                 Lang.bind(this, function () {
-                              this._select(this._currentApp,
-                                           this._currentWindows[this._currentApp]);
+                              this._select(this._currentApp, 0, true);
                               return false;
                           }));
         }
     },
 
+    _destroyThumbnails : function() {
+        Tweener.addTween(this._thumbnails.actor,
+                         { opacity: 0,
+                           time: THUMBNAIL_FADE_TIME,
+                           transition: "easeOutQuad",
+                           onComplete: function() { this.destroy(); }
+                         });
+        this._thumbnails = null;
+    },
+
     _createThumbnails : function() {
-        this._thumbnails = new ThumbnailList (this._appIcons[this._currentApp].windows);
+        this._thumbnails = new ThumbnailList (this._appIcons[this._currentApp].cachedWindows);
         this._thumbnails.connect('item-activated', Lang.bind(this, this._windowActivated));
-        this._thumbnails.connect('item-hovered', Lang.bind(this, this._windowHovered));
+        this._thumbnails.connect('item-entered', Lang.bind(this, this._windowEntered));
 
         this.actor.add_actor(this._thumbnails.actor);
 
@@ -313,12 +395,19 @@ AltTabPopup.prototype = {
             thumbnailCenter = stageX + icon.width / 2;
         } else {
             // Center the thumbnails on the monitor
-            let primary = global.get_primary_monitor();
-            thumbnailCenter = primary.x + primary.width / 2;
+            let focus = global.get_focus_monitor();
+            thumbnailCenter = focus.x + focus.width / 2;
         }
 
         this._thumbnails.actor.x = Math.floor(thumbnailCenter - this._thumbnails.actor.width / 2);
         this._thumbnails.actor.y = this._appSwitcher.actor.y + this._appSwitcher.actor.height + POPUP_LIST_SPACING;
+
+        this._thumbnails.actor.opacity = 0;
+        Tweener.addTween(this._thumbnails.actor,
+                         { opacity: 255,
+                           time: THUMBNAIL_FADE_TIME,
+                           transition: "easeOutQuad"
+                         });
     }
 };
 
@@ -329,7 +418,6 @@ function SwitcherList(squareItems) {
 SwitcherList.prototype = {
     _init : function(squareItems) {
         this.actor = new St.Bin({ style_class: 'switcher-list' });
-        this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
         // Here we use a GenericContainer so that we can force all the
         // children except the separator to have the same width.
@@ -346,93 +434,52 @@ SwitcherList.prototype = {
         this._highlighted = -1;
         this._separator = null;
         this._squareItems = squareItems;
-
-        this._hoverTimeout = 0;
-    },
-
-    _onDestroy: function() {
-        if (this._hoverTimeout != 0) {
-            Mainloop.source_remove(this._hoverTimeout);
-            this._hoverTimeout = 0;
-        }
     },
 
     addItem : function(item) {
-        let box = new St.Bin({ style_class: 'item-box' });
-        let bbox;
+        let bbox = new St.Clickable({ style_class: 'item-box',
+                                      reactive: true });
 
-        if (item instanceof Shell.ButtonBox)
-            bbox = item;
-        else {
-            bbox = new Shell.ButtonBox({ reactive: true });
-            bbox.append(item, Big.BoxPackFlags.NONE);
-        }
-        box.add_actor(bbox);
-        this._list.add_actor(box);
+        bbox.set_child(item);
+        this._list.add_actor(bbox);
 
         let n = this._items.length;
-        bbox.connect('activate', Lang.bind(this, function () {
+        bbox.connect('clicked', Lang.bind(this, function () {
                                                this._itemActivated(n);
                                           }));
-        bbox.connect('notify::hover', Lang.bind(this, function () {
-                                                    this._hoverChanged(bbox, n);
-                                               }));
+        bbox.connect('enter-event', Lang.bind(this, function () {
+                                                  this._itemEntered(n);
+                                              }));
 
-        this._items.push(box);
+        this._items.push(bbox);
     },
 
     addSeparator: function () {
-        // FIXME: make this work with StWidgets and CSS
-        let box = new Big.Box({ padding_top: 2, padding_bottom: 2 });
-        box.append(new Clutter.Rectangle({ width: 1,
-                                           color: POPUP_SEPARATOR_COLOR }),
-                   Big.BoxPackFlags.EXPAND);
+        let box = new St.Bin({ style_class: 'separator' })
         this._separator = box;
         this._list.add_actor(box);
     },
-    
-    highlight: function(index) {
+
+    highlight: function(index, justOutline) {
         if (this._highlighted != -1)
             this._items[this._highlighted].style_class = 'item-box';
 
         this._highlighted = index;
 
-        if (this._highlighted != -1)
-            this._items[this._highlighted].style_class = 'selected-item-box';
-    },
-
-    // Used after the mouse movement exceeds the threshold, to check
-    // if it's already hovering over an icon
-    checkHover: function() {
-        for (let i = 0; i < this._items.length; i++) {
-            if (this._items[i].get_child().hover) {
-                this._hoverChanged(this._items[i].get_child(), i);
-                return;
-            }
+        if (this._highlighted != -1) {
+            if (justOutline)
+                this._items[this._highlighted].style_class = 'outlined-item-box';
+            else
+                this._items[this._highlighted].style_class = 'selected-item-box';
         }
     },
 
     _itemActivated: function(n) {
         this.emit('item-activated', n);
     },
-    
-    _hoverChanged: function(box, n) {
-        if (this._hoverTimeout != 0) {
-            Mainloop.source_remove(this._hoverTimeout);
-            this._hoverTimeout = 0;
-        }
 
-        if (box.hover) {
-            this._hoverTimeout = Mainloop.timeout_add(
-                HOVER_TIME,
-                Lang.bind (this, function () {
-                               this._itemHovered(n);
-                           }));
-        }
-    },
-
-    _itemHovered: function(n) {
-        this.emit('item-hovered', n);
+    _itemEntered: function(n) {
+        this.emit('item-entered', n);
     },
 
     _maxChildWidth: function (forHeight) {
@@ -465,7 +512,7 @@ SwitcherList.prototype = {
 
         let totalSpacing = this._list.spacing * (this._items.length - 1);
         alloc.min_size = this._items.length * maxChildMin + separatorWidth + totalSpacing;
-        alloc.nat_size = this._items.length * maxChildNat + separatorWidth + totalSpacing;
+        alloc.natural_size = alloc.min_size;
     },
 
     _getPreferredHeight: function (actor, forWidth, alloc) {
@@ -481,11 +528,11 @@ SwitcherList.prototype = {
         if (this._squareItems) {
             let [childMin, childNat] = this._maxChildWidth(-1);
             maxChildMin = Math.max(childMin, maxChildMin);
-            maxChildNat = Math.max(childNat, maxChildNat);
+            maxChildNat = maxChildMin;
         }
 
         alloc.min_size = maxChildMin;
-        alloc.nat_size = maxChildNat;
+        alloc.natural_size = maxChildNat;
     },
 
     _allocate: function (actor, box, flags) {
@@ -535,6 +582,25 @@ SwitcherList.prototype = {
 
 Signals.addSignalMethods(SwitcherList.prototype);
 
+function AppIcon(app) {
+    this._init(app);
+}
+
+AppIcon.prototype = {
+    _init: function(app) {
+        this.app = app;
+        this.actor = new St.BoxLayout({ style_class: "alt-tab-app",
+                                         vertical: true });
+        this._icon = this.app.create_icon_texture(POPUP_APPICON_SIZE);
+        let iconBin = new St.Bin({height: POPUP_APPICON_SIZE, width: POPUP_APPICON_SIZE});
+        iconBin.child = this._icon;
+
+        this.actor.add(iconBin, { x_fill: false, y_fill: false } );
+        this._label = new St.Label({ text: this.app.get_name() });
+        this.actor.add(this._label, { x_fill: false });
+    }
+}
+
 function AppSwitcher(apps) {
     this._init(apps);
 }
@@ -550,8 +616,10 @@ AppSwitcher.prototype = {
         let workspaceIcons = [];
         let otherIcons = [];
         for (let i = 0; i < apps.length; i++) {
-            let appIcon = new AppIcon.AppIcon({ appInfo: apps[i],
-                                                size: POPUP_APPICON_SIZE });
+            let appIcon = new AppIcon(apps[i]);
+            // Cache the window list now; we don't handle dynamic changes here,
+            // and we don't want to be continually retrieving it
+            appIcon.cachedWindows = appIcon.app.get_windows();
             if (this._hasWindowsOnWorkspace(appIcon, activeWorkspace))
               workspaceIcons.push(appIcon);
             else
@@ -570,7 +638,7 @@ AppSwitcher.prototype = {
         for (let i = 0; i < otherIcons.length; i++)
             this._addIcon(otherIcons[i]);
 
-        this._shownArrow = -1;
+        this._curApp = -1;
     },
 
     _allocate: function (actor, box, flags) {
@@ -592,62 +660,65 @@ AppSwitcher.prototype = {
         }
     },
 
-    showArrow : function(n) {
-        if (this._shownArrow != -1)
-            this._arrows[this._shownArrow].hide();
+    // We override SwitcherList's highlight() method to also deal with
+    // the AppSwitcher->ThumbnailList arrows. Apps with only 1 window
+    // will hide their arrows by default, but show them when their
+    // thumbnails are visible (ie, when the app icon is supposed to be
+    // in justOutline mode). Apps with multiple windows will normally
+    // show a dim arrow, but show a bright arrow when they are
+    // highlighted; their redraw handler will use the right color
+    // based on this._curApp; we just need to do a queue_relayout() to
+    // force it to redraw. (queue_redraw() doesn't work because
+    // ShellDrawingArea only redraws on allocate.)
+    highlight : function(n, justOutline) {
+        if (this._curApp != -1) {
+            if (this.icons[this._curApp].cachedWindows.length == 1)
+                this._arrows[this._curApp].hide();
+            else
+                this._arrows[this._curApp].queue_relayout();
+        }
 
-        this._shownArrow = n;
+        SwitcherList.prototype.highlight.call(this, n, justOutline);
+        this._curApp = n;
 
-        if (this._shownArrow != -1)
-            this._arrows[this._shownArrow].show();
+        if (this._curApp != -1) {
+            if (justOutline && this.icons[this._curApp].cachedWindows.length == 1)
+                this._arrows[this._curApp].show();
+            else
+                this._arrows[this._curApp].queue_relayout();
+        }
     },
 
     _addIcon : function(appIcon) {
         this.icons.push(appIcon);
         this.addItem(appIcon.actor);
 
-        let arrow = new Shell.DrawingArea();
+        let n = this._arrows.length;
+        let arrow = new St.DrawingArea();
         arrow.connect('redraw', Lang.bind(this,
             function (area, texture) {
                 Shell.draw_box_pointer(texture, Shell.PointerDirection.DOWN,
                                        TRANSPARENT_COLOR,
-                                       POPUP_ARROW_COLOR);
+                                       this._curApp == n ? POPUP_ARROW_COLOR : POPUP_UNFOCUSED_ARROW_COLOR);
             }));
         this._list.add_actor(arrow);
         this._arrows.push(arrow);
-        arrow.hide();
+
+        if (appIcon.cachedWindows.length == 1)
+            arrow.hide();
     },
 
     _hasWindowsOnWorkspace: function(appIcon, workspace) {
-        for (let i = 0; i < appIcon.windows.length; i++) {
-            if (appIcon.windows[i].get_workspace() == workspace)
-                return true;
-        }
-        return false;
-    },
-
-    _hasVisibleWindows : function(appIcon) {
-        for (let i = 0; i < appIcon.windows.length; i++) {
-            if (appIcon.windows[i].showing_on_its_workspace())
+        let windows = appIcon.cachedWindows;
+        for (let i = 0; i < windows.length; i++) {
+            if (windows[i].get_workspace() == workspace)
                 return true;
         }
         return false;
     },
 
     _sortAppIcon : function(appIcon1, appIcon2) {
-        let vis1 = this._hasVisibleWindows(appIcon1);
-        let vis2 = this._hasVisibleWindows(appIcon2);
-
-        if (vis1 && !vis2) {
-            return -1;
-        } else if (vis2 && !vis1) {
-            return 1;
-        } else {
-            // The app's most-recently-used window is first
-            // in its list
-            return (appIcon2.windows[0].get_user_time() -
-                    appIcon1.windows[0].get_user_time());
-        }
+        return appIcon1.app.compare(appIcon2.app);
     }
 };
 
@@ -661,18 +732,45 @@ ThumbnailList.prototype = {
     _init : function(windows) {
         SwitcherList.prototype._init.call(this);
 
+        let activeWorkspace = global.screen.get_active_workspace();
+
+        // We fake the value of "separatorAdded" when the app has no window
+        // on the current workspace, to avoid displaying a useless separator in
+        // that case.
+        let separatorAdded = windows.length == 0 || windows[0].get_workspace() != activeWorkspace;
+
         for (let i = 0; i < windows.length; i++) {
+            if (!separatorAdded && windows[i].get_workspace() != activeWorkspace) {
+              this.addSeparator();
+              separatorAdded = true;
+            }
+
             let mutterWindow = windows[i].get_compositor_private();
             let windowTexture = mutterWindow.get_texture ();
             let [width, height] = windowTexture.get_size();
             let scale = Math.min(1.0, THUMBNAIL_SIZE / width, THUMBNAIL_SIZE / height);
 
+            let box = new St.BoxLayout({ style_class: "thumbnail-box",
+                                         vertical: true });
+
+            let bin = new St.Bin({ style_class: "thumbnail" });
             let clone = new Clutter.Clone ({ source: windowTexture,
                                              reactive: true,
                                              width: width * scale,
                                              height: height * scale });
-            let box = new Big.Box({ padding: AppIcon.APPICON_BORDER_WIDTH + AppIcon.APPICON_PADDING });
-            box.append(clone, Big.BoxPackFlags.NONE);
+
+            bin.add_actor(clone);
+            box.add_actor(bin);
+
+            let title = windows[i].get_title();
+            if (title) {
+                let name = new St.Label({ text: title });
+                // St.Label doesn't support text-align so use a Bin
+                let bin = new St.Bin({ x_align: St.Align.MIDDLE });
+                bin.add_actor(name);
+                box.add_actor(bin);
+            }
+
             this.addItem(box);
         }
     }
