@@ -8,13 +8,18 @@ const Lang = imports.lang;
 const Pango = imports.gi.Pango;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
+const St = imports.gi.St;
 const Mainloop = imports.mainloop;
+const Gettext = imports.gettext.domain('gnome-shell');
+const _ = Gettext.gettext;
 
 const DocInfo = imports.misc.docInfo;
 const DND = imports.ui.dnd;
 const GenericDisplay = imports.ui.genericDisplay;
 const Main = imports.ui.main;
+const Search = imports.ui.search;
 
+const MAX_DASH_DOCS = 50;
 const DASH_DOCS_ICON_SIZE = 16;
 
 const DEFAULT_SPACING = 4;
@@ -111,19 +116,19 @@ DocDisplayItem.prototype = {
 /* This class represents a display containing a collection of document items.
  * The documents are sorted by how recently they were last visited.
  */
-function DocDisplay() {
-    this._init();
+function DocDisplay(flags) {
+    this._init(flags);
 }
 
 DocDisplay.prototype = {
     __proto__:  GenericDisplay.GenericDisplay.prototype,
 
-    _init : function() {
-        GenericDisplay.GenericDisplay.prototype._init.call(this);
+    _init : function(flags) {
+        GenericDisplay.GenericDisplay.prototype._init.call(this, flags);
         // We keep a single timeout callback for updating last visited times
         // for all the items in the display. This avoids creating individual
         // callbacks for each item in the display. So proper time updates
-        // for individual items and item details depend on the item being 
+        // for individual items and item details depend on the item being
         // associated with one of the displays.
         this._updateTimeoutTargetTime = -1;
         this._updateTimeoutId = 0;
@@ -151,7 +156,8 @@ DocDisplay.prototype = {
     _refreshCache : function() {
         if (!this._docsStale)
             return true;
-        this._allItems = this._docManager.getItems();
+        this._allItems = {};
+        Lang.copyProperties(this._docManager.getInfosByUri(), this._allItems);
         this._docsStale = false;
         return false;
     },
@@ -176,13 +182,8 @@ DocDisplay.prototype = {
         this._matchedItemKeys = [];
         let docIdsToRemove = [];
         for (docId in this._allItems) {
-            // this._allItems[docId].exists() checks if the resource still exists
-            if (this._allItems[docId].exists()) {
-                this._matchedItems[docId] = 1;
-                this._matchedItemKeys.push(docId);
-            } else {
-                docIdsToRemove.push(docId);
-            }
+            this._matchedItems[docId] = 1;
+            this._matchedItemKeys.push(docId);
         }
 
         for (docId in docIdsToRemove) {
@@ -274,18 +275,21 @@ DashDocDisplayItem.prototype = {
             Main.overview.hide();
         }));
 
+        this.actor._delegate = this;
+
         this._icon = docInfo.createIcon(DASH_DOCS_ICON_SIZE);
         let iconBox = new Big.Box({ y_align: Big.BoxAlignment.CENTER });
         iconBox.append(this._icon, Big.BoxPackFlags.NONE);
         this.actor.append(iconBox, Big.BoxPackFlags.NONE);
-        let name = new Clutter.Text({ font_name: "Sans 14px",
-                                      color: GenericDisplay.ITEM_DISPLAY_NAME_COLOR,
-                                      ellipsize: Pango.EllipsizeMode.END,
-                                      text: docInfo.name });
+        let name = new St.Label({ style_class: 'dash-recent-docs-item',
+                                   text: docInfo.name });
         this.actor.append(name, Big.BoxPackFlags.EXPAND);
 
         let draggable = DND.makeDraggable(this.actor);
-        this.actor._delegate = this;
+    },
+
+    getUri: function() {
+        return this._info.uri;
     },
 
     getDragActorSource: function() {
@@ -317,12 +321,14 @@ DashDocDisplay.prototype = {
         this.actor.connect('get-preferred-width', Lang.bind(this, this._getPreferredWidth));
         this.actor.connect('get-preferred-height', Lang.bind(this, this._getPreferredHeight));
         this.actor.connect('allocate', Lang.bind(this, this._allocate));
+        this._workId = Main.initializeDeferredWork(this.actor, Lang.bind(this, this._redisplay));
+
+        this._actorsByUri = {};
 
         this._docManager = DocInfo.getDocManager();
-        this._docManager.connect('changed', Lang.bind(this, function(mgr) {
-            this._redisplay();
-        }));
-        this._redisplay();
+        this._docManager.connect('changed', Lang.bind(this, this._onDocsChanged));
+        this._pendingDocsChange = true;
+        this._checkDocExistence = false;
     },
 
     _getPreferredWidth: function(actor, forHeight, alloc) {
@@ -351,90 +357,119 @@ DashDocDisplay.prototype = {
     _getPreferredHeight: function(actor, forWidth, alloc) {
         let children = actor.get_children();
 
-        // Two columns, where we go vertically down first.  So just take
-        // the height of half of the children as our preferred height.
+        // The width of an item is our allocated width, minus spacing, divided in half.
+        this._itemWidth = Math.floor((forWidth - DEFAULT_SPACING) / 2);
 
-        let firstColumnChildren = children.length / 2;
-
-        alloc.min_size = 0;
-        for (let i = 0; i < firstColumnChildren; i++) {
+        let maxNatural = 0;
+        for (let i = 0; i < children.length; i++) {
             let child = children[i];
-            let [minSize, naturalSize] = child.get_preferred_height(forWidth);
-            alloc.natural_size += naturalSize;
-
-            if (i > 0 && i < children.length - 1) {
-                alloc.min_size += DEFAULT_SPACING;
-                alloc.natural_size += DEFAULT_SPACING;
-            }
+            let [minSize, naturalSize] = child.get_preferred_height(this._itemWidth);
+            maxNatural = Math.max(maxNatural, naturalSize);
         }
+
+        this._itemHeight = maxNatural;
+
+        let firstColumnChildren = Math.ceil(children.length / 2);
+        alloc.natural_size = (firstColumnChildren * maxNatural +
+                              (firstColumnChildren - 1) * DEFAULT_SPACING);
     },
 
     _allocate: function(actor, box, flags) {
         let width = box.x2 - box.x1;
         let height = box.y2 - box.y1;
 
+        // Make sure this._itemWidth/Height have been computed, even
+        // if the parent actor didn't check our size before allocating.
+        // (Not clear if that is required or not as a Clutter
+        // invariant; this is safe and cheap because of caching.)
+        actor.get_preferred_height(width);
+
         let children = actor.get_children();
 
-        // The width of an item is our allocated width, minus spacing, divided in half.
-        let itemWidth = Math.floor((width - DEFAULT_SPACING) / 2);
-        let x = box.x1;
-        let y = box.y1;
+        let x = 0;
+        let y = 0;
         let columnIndex = 0;
         let i = 0;
         // Loop over the children, going vertically down first.  When we run
         // out of vertical space (our y variable is bigger than box.y2), switch
         // to the second column.
-        for (; i < children.length; i++) {
+        while (i < children.length) {
             let child = children[i];
 
-            let [minSize, naturalSize] = child.get_preferred_height(-1);
-
-            if (y + naturalSize > box.y2) {
-                // Is this the second column?  Ok, break.
-                if (columnIndex == 1) {
+            if (y + this._itemHeight > box.y2) {
+                // Is this the second column, or we're in
+                // the first column and can't even fit one
+                // item?  In that case, break.
+                if (columnIndex == 1 || i == 0) {
                     break;
                 }
                 // Set x to the halfway point.
                 columnIndex += 1;
-                x = x + itemWidth + DEFAULT_SPACING;
+                x = x + this._itemWidth + DEFAULT_SPACING;
                 // And y is back to the top.
-                y = box.y1;
+                y = 0;
+                // Retry this same item, now that we're in the second column.
+                // By looping back to the top here, we re-test the size
+                // again for the second column.
+                continue;
             }
 
             let childBox = new Clutter.ActorBox();
             childBox.x1 = x;
             childBox.y1 = y;
-            childBox.x2 = childBox.x1 + itemWidth;
-            childBox.y2 = y + naturalSize;
+            childBox.x2 = childBox.x1 + this._itemWidth;
+            childBox.y2 = y + this._itemHeight;
 
             y = childBox.y2 + DEFAULT_SPACING;
 
-            child.show();
             child.allocate(childBox, flags);
+            this.actor.set_skip_paint(child, false);
+
+            i++;
         }
 
-        // Everything else didn't fit, just hide it.
-        for (; i < children.length; i++) {
-            children[i].hide();
+        if (this._checkDocExistence) {
+            // Now we know how many docs we are displaying, queue a check to see if any of them
+            // have been deleted. If they are deleted, then we'll get a 'changed' signal; since
+            // we'll now be displaying items we weren't previously, we'll check again to see
+            // if they were deleted, and so forth and so on.
+            // TODO: We should change this to ask for as many as we can fit in the given space:
+            // https://bugzilla.gnome.org/show_bug.cgi?id=603522#c23
+            this._docManager.queueExistenceCheck(i);
+            this._checkDocExistence = false;
         }
+
+        for (; i < children.length; i++)
+            this.actor.set_skip_paint(children[i], true);
+    },
+
+    _onDocsChanged: function() {
+        this._checkDocExistence = true;
+        Main.queueDeferredWork(this._workId);
     },
 
     _redisplay: function() {
+        // Should be kept alive by the _actorsByUri
         this.actor.remove_all();
-
-        let docs = this._docManager.getItems();
-        let docUrls = [];
-        for (let url in docs) {
-            docUrls.push(url);
+        let docs = this._docManager.getTimestampOrderedInfos();
+        for (let i = 0; i < docs.length && i < MAX_DASH_DOCS; i++) {
+            let doc = docs[i];
+            let display = this._actorsByUri[doc.uri];
+            if (display) {
+                this.actor.add_actor(display.actor);
+            } else {
+                let display = new DashDocDisplayItem(doc);
+                this.actor.add_actor(display.actor);
+                this._actorsByUri[doc.uri] = display;
+            }
         }
-        docUrls.sort(function (urlA, urlB) { return docs[urlB].timestamp - docs[urlA].timestamp; });
-        let textureCache = Shell.TextureCache.get_default();
-
-        for (let i = 0; i < docUrls.length; i++) {
-            let url = docUrls[i];
-            let docInfo = docs[url];
-            let display = new DashDocDisplayItem(docInfo);
-            this.actor.add_actor(display.actor);
+        // Any unparented actors must have been deleted
+        for (let uri in this._actorsByUri) {
+            let display = this._actorsByUri[uri];
+            if (display.actor.get_parent() == null) {
+                display.actor.destroy();
+                delete this._actorsByUri[uri];
+            }
         }
         this.emit('changed');
     }
@@ -442,3 +477,41 @@ DashDocDisplay.prototype = {
 
 Signals.addSignalMethods(DashDocDisplay.prototype);
 
+function DocSearchProvider() {
+    this._init();
+}
+
+DocSearchProvider.prototype = {
+    __proto__: Search.SearchProvider.prototype,
+
+    _init: function(name) {
+        Search.SearchProvider.prototype._init.call(this, _("RECENT ITEMS"));
+        this._docManager = DocInfo.getDocManager();
+    },
+
+    getResultMeta: function(resultId) {
+        let docInfo = this._docManager.lookupByUri(resultId);
+        if (!docInfo)
+            return null;
+        return { 'id': resultId,
+                 'name': docInfo.name,
+                 'icon': docInfo.createIcon(Search.RESULT_ICON_SIZE)};
+    },
+
+    activateResult: function(id) {
+        let docInfo = this._docManager.lookupByUri(id);
+        docInfo.launch();
+    },
+
+    getInitialResultSet: function(terms) {
+        return this._docManager.initialSearch(terms);
+    },
+
+    getSubsearchResultSet: function(previousResults, terms) {
+        return this._docManager.subsearch(previousResults, terms);
+    },
+
+    expandSearch: function(terms) {
+        log("TODO expand docs search");
+    }
+};
