@@ -8,19 +8,18 @@ const Mainloop = imports.mainloop;
 const Shell = imports.gi.Shell;
 const Signals = imports.signals;
 const Lang = imports.lang;
+const St = imports.gi.St;
+const Gettext = imports.gettext.domain('gnome-shell');
+const _ = Gettext.gettext;
 
 const AppDisplay = imports.ui.appDisplay;
 const DocDisplay = imports.ui.docDisplay;
 const GenericDisplay = imports.ui.genericDisplay;
-const Link = imports.ui.link;
 const Main = imports.ui.main;
 const Panel = imports.ui.panel;
 const Dash = imports.ui.dash;
 const Tweener = imports.ui.tweener;
-const Workspaces = imports.ui.workspaces;
-
-const ROOT_OVERVIEW_COLOR = new Clutter.Color();
-ROOT_OVERVIEW_COLOR.from_pixel(0x000000ff);
+const WorkspacesView = imports.ui.workspacesView;
 
 // Time for initial animation going into Overview mode
 const ANIMATION_TIME = 0.25;
@@ -74,10 +73,117 @@ const SHADOW_WIDTH = 6;
 
 const NUMBER_OF_SECTIONS_IN_SEARCH = 2;
 
+const INFO_BAR_HIDE_TIMEOUT = 30;
+
 let wideScreen = false;
 let displayGridColumnWidth = null;
 let displayGridRowHeight = null;
-let addRemoveButtonSize = null;
+
+function InfoBar() {
+    this._init();
+}
+
+InfoBar.prototype = {
+    _init: function() {
+        this.actor = new St.Bin({ style_class: 'info-bar-panel',
+                                  x_fill: true,
+                                  y_fill: false });
+        this._label = new St.Label();
+        this._undo = new St.Button({ label: _("Undo"),
+                                     style_class: 'info-bar-link-button' });
+
+        let bin = new St.Bin({ x_fill: false,
+                               y_fill: false,
+                               x_align: St.Align.MIDDLE,
+                               y_align: St.Align.MIDDLE });
+        this.actor.set_child(bin);
+
+        let box = new St.BoxLayout({ style_class: 'info-bar' });
+        bin.set_child(box);
+        this._timeoutId = 0;
+
+        box.add(this._label, {'y-fill' : false, 'y-align' : St.Align.MIDDLE});
+        box.add(this._undo);
+
+        this.actor.set_opacity(0);
+
+        this._undoCallback = null;
+        this._undo.connect('clicked', Lang.bind(this, this._onUndoClicked));
+
+        this._overviewWasHidden = false;
+        this._hidingOverviewId = 0;
+    },
+
+    _onUndoClicked: function() {
+        Mainloop.source_remove(this._timeoutId);
+        this._timeoutId = 0;
+
+        if (this._undoCallback)
+            this._undoCallback();
+        this.actor.set_opacity(0);
+        this._undoCallback = null;
+    },
+
+    _hideDone: function() {
+        this._undoCallback = null;
+    },
+
+    _hide: function() {
+        this._overviewWasHidden = false;
+        Tweener.addTween(this.actor,
+                         { opacity: 0,
+                           transition: 'easeOutQuad',
+                           time: ANIMATION_TIME,
+                           onComplete: this._hideDone,
+                           onCompleteScope: this
+                         });
+    },
+
+    _onTimeout: function() {
+        this._timeoutId = 0;
+        if (this._overviewWasHidden)
+            this._hide();
+        return false;
+    },
+
+    _onOverviewHiding: function() {
+        if (this._timeoutId == 0)
+            this._hide();
+        else
+            this._overviewWasHidden = true;
+    },
+
+    setMessage: function(text, undoCallback) {
+        if (this._timeoutId)
+            Mainloop.source_remove(this._timeoutId);
+
+        if (this._hidingOverviewId == 0) {
+            // Set here, because when constructor is called, overview is null.
+            if (!Main.overview)
+                return;
+            // We don't actually use the ID, it's just a way of tracking whether we've hooked up the signal
+            this._hidingOverviewId = Main.overview.connect('hiding', Lang.bind(this, this._onOverviewHiding));
+        }
+        this._timeout = false;
+        this._overviewWasHidden = false;
+
+        this._label.text = text;
+
+        Tweener.addTween(this.actor,
+                         { opacity: 255,
+                           transition: 'easeOutQuad',
+                           time: ANIMATION_TIME
+                         });
+
+        this._timeoutId = Mainloop.timeout_add_seconds(INFO_BAR_HIDE_TIMEOUT, Lang.bind(this, this._onTimeout));
+
+        this._undoCallback = undoCallback;
+        if (undoCallback)
+            this._undo.show();
+        else
+            this._undo.hide();
+    }
+};
 
 function Overview() {
     this._init();
@@ -85,9 +191,15 @@ function Overview() {
 
 Overview.prototype = {
     _init : function() {
-        this._group = new Clutter.Group();
+        this._group = new St.BoxLayout({ style_class: 'overview' });
         this._group._delegate = this;
 
+        this.infoBar = new InfoBar();
+        this._group.add_actor(this.infoBar.actor);
+
+        this._workspacesViewSwitch = new WorkspacesView.WorkspacesViewSwitch();
+        this._workspacesViewSwitch.connect('view-changed', Lang.bind(this, this._onViewChanged));
+
         this.visible = false;
         this.animationInProgress = false;
         this._hideInProgress = false;
@@ -111,7 +223,7 @@ Overview.prototype = {
         this._group.add_actor(this._transparentBackground);
 
         // Background color for the Overview
-        this._backOver = new Clutter.Rectangle({ color: ROOT_OVERVIEW_COLOR });
+        this._backOver = new St.Label();
         this._group.add_actor(this._backOver);
 
         this._group.hide();
@@ -122,9 +234,7 @@ Overview.prototype = {
         this._group.add_actor(this._dash.actor);
 
         // Container to hold popup pane chrome.
-        this._paneContainer = new Big.Box({ orientation: Big.BoxOrientation.HORIZONTAL,
-                                            spacing: 6
-                                          });
+        this._paneContainer = new St.BoxLayout({ style_class: 'overview-pane' });
         // Note here we explicitly don't set the paneContainer to be reactive yet; that's done
         // inside the notify::visible handler on panes.
         this._paneContainer.connect('button-release-event', Lang.bind(this, function(background) {
@@ -134,13 +244,52 @@ Overview.prototype = {
         this._group.add_actor(this._paneContainer);
 
         this._transparentBackground.lower_bottom();
-        this._paneContainer.lower_bottom();
+        this._paneContainer.hide();
 
         this._coverPane.lower_bottom();
 
         this._workspaces = null;
     },
 
+    _createControlsBar: function() {
+        this._workspacesBar = new St.BoxLayout({ 'pack-start': true,
+                                                 style_class: 'workspaces-bar' });
+        this._workspacesBar.move_by(this._workspacesBarX, this._workspacesBarY);
+
+        let controlsBar = this._workspacesViewSwitch.createControlsBar();
+        let bar = this._workspaces.createControllerBar();
+        this._workspacesBar.add(bar, { expand: true, 'x-fill': true, 'y-fill': true,
+                                       y_align: St.Align.MIDDLE, x_align: St.Align.START });
+        this._workspacesBar.add(controlsBar, {x_align: St.Align.END});
+        this._workspacesBar.width = this._workspacesBarWidth;
+
+        this._group.add_actor(this._workspacesBar);
+        this._workspacesBar.raise(this._workspaces.actor);
+    },
+
+    _onViewChanged: function() {
+        if (!this.visible)
+            return;
+        //Remove old worspacesView
+        this._group.remove_actor(this._workspacesBar);
+        this._workspaces.hide();
+        this._group.remove_actor(this._workspaces.actor);
+        this._workspaces.destroy();
+        this._workspacesBar.destroy();
+
+        this._workspaces = this._workspacesViewSwitch.createCurrentWorkspaceView(this._workspacesWidth, this._workspacesHeight,
+                                                                             this._workspacesX, this._workspacesY, false);
+
+        //Show new workspacesView
+        this._group.add_actor(this._workspaces.actor);
+        this._dash.actor.raise(this._workspaces.actor);
+
+        this._createControlsBar();
+
+        // Set new position and scale to workspaces.
+        this.emit('showing');
+    },
+
     _recalculateGridSizes: function () {
         let primary = global.get_primary_monitor();
         wideScreen = (primary.width/primary.height > WIDE_SCREEN_CUT_OFF_RATIO) &&
@@ -149,17 +298,19 @@ Overview.prototype = {
         // We divide the screen into an imaginary grid which helps us determine the layout of
         // different visual components.
         if (wideScreen) {
-            displayGridColumnWidth = primary.width / COLUMNS_WIDE_SCREEN;
-            displayGridRowHeight = primary.height / ROWS_WIDE_SCREEN;
+            displayGridColumnWidth = Math.floor(primary.width / COLUMNS_WIDE_SCREEN);
+            displayGridRowHeight = Math.floor(primary.height / ROWS_WIDE_SCREEN);
         } else {
-            displayGridColumnWidth = primary.width / COLUMNS_REGULAR_SCREEN;
-            displayGridRowHeight = primary.height / ROWS_REGULAR_SCREEN;
+            displayGridColumnWidth = Math.floor(primary.width / COLUMNS_REGULAR_SCREEN);
+            displayGridRowHeight = Math.floor(primary.height / ROWS_REGULAR_SCREEN);
         }
     },
 
     relayout: function () {
         let primary = global.get_primary_monitor();
 
+        this._recalculateGridSizes();
+
         this._group.set_position(primary.x, primary.y);
 
         let contentY = Panel.PANEL_HEIGHT;
@@ -175,21 +326,26 @@ Overview.prototype = {
                                   - WORKSPACE_GRID_PADDING * 2;
         // We scale the vertical padding by (primary.height / primary.width)
         // so that the workspace preserves its aspect ratio.
-        this._workspacesHeight = displayGridRowHeight * workspaceRowsUsed
-                                   - WORKSPACE_GRID_PADDING * (primary.height / primary.width) * 2;
+        this._workspacesHeight = Math.floor(displayGridRowHeight * workspaceRowsUsed
+                                   - WORKSPACE_GRID_PADDING * (primary.height / primary.width) * 2);
 
         this._workspacesX = displayGridColumnWidth + WORKSPACE_GRID_PADDING;
-        this._workspacesY = displayGridRowHeight + WORKSPACE_GRID_PADDING * (primary.height / primary.width);
+        this._workspacesY = Math.floor(displayGridRowHeight + WORKSPACE_GRID_PADDING * (primary.height / primary.width));
 
         this._dash.actor.set_position(0, contentY);
         this._dash.actor.set_size(displayGridColumnWidth, contentHeight);
         this._dash.searchArea.height = this._workspacesY - contentY;
         this._dash.sectionArea.height = this._workspacesHeight;
+        this._dash.searchResults.actor.height = this._workspacesHeight;
+
+        this.infoBar.actor.set_position(displayGridColumnWidth, Panel.PANEL_HEIGHT);
+        this.infoBar.actor.set_size(primary.width - displayGridColumnWidth, this._workspacesY - Panel.PANEL_HEIGHT);
+        this.infoBar.actor.raise_top();
 
         // place the 'Add Workspace' button in the bottom row of the grid
-        addRemoveButtonSize = Math.floor(displayGridRowHeight * 3/5);
-        this._addButtonX = this._workspacesX + this._workspacesWidth - addRemoveButtonSize;
-        this._addButtonY = primary.height - Math.floor(displayGridRowHeight * 4/5);
+        this._workspacesBarX = this._workspacesX;
+        this._workspacesBarWidth = primary.width - this._workspacesBarX - WORKSPACE_GRID_PADDING;
+        this._workspacesBarY = primary.height - displayGridRowHeight;
 
         // The parent (this._group) is positioned at the top left of the primary monitor
         // while this._backOver occupies the entire screen.
@@ -197,20 +353,18 @@ Overview.prototype = {
         this._backOver.set_size(global.screen_width, global.screen_height);
 
         this._paneContainer.set_position(this._dash.actor.x + this._dash.actor.width + DEFAULT_PADDING,
-                                         contentY);
+                                         this._workspacesY);
         // Dynamic width
-        this._paneContainer.height = contentHeight;
+        this._paneContainer.height = this._workspacesHeight;
 
         this._transparentBackground.set_position(this._paneContainer.x, this._paneContainer.y);
         this._transparentBackground.set_size(primary.width - this._paneContainer.x,
                                              this._paneContainer.height);
 
-        if (this._activeDisplayPane != null)
-            this._activeDisplayPane.actor.width = displayGridColumnWidth * 2;
     },
 
     addPane: function (pane) {
-        this._paneContainer.append(pane.actor, Big.BoxPackFlags.NONE);
+        this._paneContainer.add(pane.actor, { expand: true, y_fill: false, y_align: St.Align.START });
         // When a pane is displayed, we raise the transparent background to the top
         // and connect to button-release-event on it, then raise the pane above that.
         // The idea here is that clicking anywhere outside the pane should close it.
@@ -218,16 +372,17 @@ Overview.prototype = {
         let backgroundEventId = null;
         pane.connect('open-state-changed', Lang.bind(this, function (pane, isOpen) {
             if (isOpen) {
-                pane.actor.width = displayGridColumnWidth * 2;
                 this._activeDisplayPane = pane;
                 this._transparentBackground.raise_top();
                 this._paneContainer.raise_top();
+                this._paneContainer.show();
                 if (backgroundEventId != null)
                     this._transparentBackground.disconnect(backgroundEventId);
                 backgroundEventId = this._transparentBackground.connect('button-release-event', Lang.bind(this, function () {
                     this._activeDisplayPane.close();
                     return true;
                 }));
+                this._workspaces.actor.opacity = 64;
             } else if (pane == this._activeDisplayPane) {
                 this._activeDisplayPane = null;
                 if (backgroundEventId != null) {
@@ -235,7 +390,8 @@ Overview.prototype = {
                     backgroundEventId = null;
                 }
                 this._transparentBackground.lower_bottom();
-                this._paneContainer.lower_bottom();
+                this._paneContainer.hide();
+                this._workspaces.actor.opacity = 255;
             }
         }));
     },
@@ -247,7 +403,7 @@ Overview.prototype = {
     // This allows the user to place the item on any workspace.
     handleDragOver : function(source, actor, x, y, time) {
         if (source instanceof GenericDisplay.GenericDisplayItem
-            || source instanceof AppDisplay.BaseWellItem) {
+            || source instanceof AppDisplay.AppIcon) {
             if (this._activeDisplayPane != null)
                 this._activeDisplayPane.close();
             return true;
@@ -295,8 +451,8 @@ Overview.prototype = {
         this._dash.show();
 
         /* TODO: make this stuff dynamic */
-        this._workspaces = new Workspaces.Workspaces(this._workspacesWidth, this._workspacesHeight,
-                                                     this._workspacesX, this._workspacesY);
+        this._workspaces = this._workspacesViewSwitch.createCurrentWorkspaceView(this._workspacesWidth, this._workspacesHeight,
+                                                                             this._workspacesX, this._workspacesY, true);
         this._group.add_actor(this._workspaces.actor);
 
         // The workspaces actor is as big as the screen, so we have to raise the dash above it
@@ -304,11 +460,7 @@ Overview.prototype = {
         // be as big as the screen.
         this._dash.actor.raise(this._workspaces.actor);
 
-        // Create (+) button
-        this._addButton = new AddWorkspaceButton(addRemoveButtonSize, this._addButtonX, this._addButtonY, Lang.bind(this, this._acceptNewWorkspaceDrop));
-        this._addButton.actor.connect('button-release-event', Lang.bind(this, this._addNewWorkspace));
-        this._group.add_actor(this._addButton.actor);
-        this._addButton.actor.raise(this._workspaces.actor);
+        this._createControlsBar();
 
         // All the the actors in the window group are completely obscured,
         // hiding the group holding them while the Overview is displayed greatly
@@ -360,9 +512,8 @@ Overview.prototype = {
             this._activeDisplayPane.close();
         this._workspaces.hide();
 
-        this._addButton.actor.destroy();
-        this._addButton.actor = null;
-        this._addButton = null;
+        this._workspacesBar.destroy();
+        this._workspacesBar = null;
 
         // Create a zoom in effect by transforming the Overview group so that
         // the active workspace fills up the whole screen. The opposite
@@ -410,20 +561,6 @@ Overview.prototype = {
         return this._workspaces;
     },
 
-    /**
-     * activateWindow:
-     * @metaWindow: A #MetaWindow
-     * @time: Event timestamp integer
-     *
-     * Make the given MetaWindow be the focus window, switching
-     * to the workspace it's on if necessary.  This function
-     * should only be used when the Overview is currently active;
-     * outside of that, use the relevant methods on MetaDisplay.
-     */
-    activateWindow: function (metaWindow, time) {
-         this._workspaces.activateWindowFromOverview(metaWindow, time);
-    },
-
     //// Private methods ////
 
     _showDone: function() {
@@ -445,7 +582,7 @@ Overview.prototype = {
         this._dash.hide();
         this._group.hide();
 
-        this.visible = false; 
+        this.visible = false;
         this.animationInProgress = false;
         this._hideInProgress = false;
 
@@ -453,43 +590,6 @@ Overview.prototype = {
 
         Main.popModal(this._dash.actor);
         this.emit('hidden');
-    },
-
-    _addNewWorkspace: function() {
-        global.screen.append_new_workspace(false, Main.currentTime());
-    },
-
-    _acceptNewWorkspaceDrop: function(source, dropActor, x, y, time) {
-        this._addNewWorkspace();
-        return this._workspaces.acceptNewWorkspaceDrop(source, dropActor, x, y, time);
     }
 };
 Signals.addSignalMethods(Overview.prototype);
-
-function AddWorkspaceButton(buttonSize, buttonX, buttonY, acceptDropCallback) {
-    this._init(buttonSize, buttonX, buttonY, acceptDropCallback);
-}
-
-AddWorkspaceButton.prototype = {
-    _init: function(buttonSize, buttonX, buttonY, acceptDropCallback) {
-        this.actor = new Clutter.Group({ x: buttonX,
-                                         y: buttonY,
-                                         width: global.screen_width - buttonX,
-                                         height: global.screen_height - buttonY,
-                                         reactive: true });
-        this.actor._delegate = this;
-        this._acceptDropCallback = acceptDropCallback;
-
-        let plus = new Clutter.Texture({ x: 0,
-                                         y: 0,
-                                         width: buttonSize,
-                                         height: buttonSize });
-        plus.set_from_file(global.imagedir + 'add-workspace.svg');
-        this.actor.add_actor(plus);
-    },
-
-    // Draggable target interface
-    acceptDrop: function(source, actor, x, y, time) {
-        return this._acceptDropCallback(source, actor, x, y, time);
-    }
-};
