@@ -1,9 +1,11 @@
+/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
 #ifndef __SHELL_APP_SYSTEM_H__
 #define __SHELL_APP_SYSTEM_H__
 
 #include <gio/gio.h>
 #include <clutter/clutter.h>
 
+#include "shell-app.h"
 #include "window.h"
 
 #define SHELL_TYPE_APP_SYSTEM                 (shell_app_system_get_type ())
@@ -35,18 +37,6 @@ struct _ShellAppSystemClass
 GType shell_app_system_get_type (void) G_GNUC_CONST;
 ShellAppSystem* shell_app_system_get_default(void);
 
-GSList *shell_app_system_get_applications_for_menu (ShellAppSystem *system, const char *menu);
-
-typedef struct _ShellAppMenuEntry ShellAppMenuEntry;
-
-struct _ShellAppMenuEntry {
-  char *name;
-  char *id;
-  char *icon;
-};
-
-GType shell_app_menu_entry_get_type (void);
-
 typedef struct _ShellAppInfo ShellAppInfo;
 
 #define SHELL_TYPE_APP_INFO (shell_app_info_get_type ())
@@ -75,20 +65,26 @@ gboolean shell_app_info_launch (ShellAppInfo *info,
 
 ShellAppInfo *shell_app_system_load_from_desktop_file (ShellAppSystem *system, const char *filename, GError **error);
 
-ShellAppInfo *shell_app_system_lookup_cached_app (ShellAppSystem *system, const char *id);
+ShellApp *shell_app_system_get_app (ShellAppSystem *system, const char *id);
+ShellApp *shell_app_system_get_app_for_window (ShellAppSystem *self, MetaWindow *window);
 
-ShellAppInfo *shell_app_system_lookup_heuristic_basename (ShellAppSystem *system, const char *id);
+void _shell_app_system_register_app (ShellAppSystem *self, ShellApp *app);
+
+ShellApp *shell_app_system_lookup_heuristic_basename (ShellAppSystem *system, const char *id);
 
 ShellAppInfo *shell_app_system_create_from_window (ShellAppSystem *system, MetaWindow *window);
 
-GSList *shell_app_system_get_menus (ShellAppSystem *system);
+GSList *shell_app_system_get_flattened_apps (ShellAppSystem *system);
 
 GSList *shell_app_system_get_all_settings (ShellAppSystem *system);
 
-GList *shell_app_system_get_favorites (ShellAppSystem *system);
+GSList *shell_app_system_initial_search (ShellAppSystem *system,
+                                         gboolean        prefs,
+                                         GSList         *terms);
 
-void shell_app_system_add_favorite (ShellAppSystem *system, const char *id);
-
-void shell_app_system_remove_favorite (ShellAppSystem *system, const char *id);
+GSList *shell_app_system_subsearch (ShellAppSystem   *system,
+                                    gboolean          prefs,
+                                    GSList           *previous_results,
+                                    GSList           *terms);
 
 #endif /* __SHELL_APP_SYSTEM_H__ */
