@@ -1,104 +1,10 @@
 /* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
 
+#include "config.h"
+
 #include "shell-drawing.h"
 #include <math.h>
 
-/**
- * shell_create_vertical_gradient:
- * @top: the color at the top
- * @bottom: the color at the bottom
- *
- * Creates a vertical gradient actor.
- *
- * Return value: (transfer none): a #ClutterCairoTexture actor with the
- *               gradient. The texture actor is floating, hence (transfer none).
- */
-ClutterCairoTexture *
-shell_create_vertical_gradient (ClutterColor *top,
-                                ClutterColor *bottom)
-{
-  ClutterCairoTexture *texture;
-  cairo_t *cr;
-  cairo_pattern_t *pattern;
-
-  /* Draw the gradient on an 8x8 pixel texture. Because the gradient is drawn
-   * from the uppermost to the lowermost row, after stretching 1/16 of the
-   * texture height has the top color and 1/16 has the bottom color. The 8
-   * pixel width is chosen for reasons related to graphics hardware internals.
-   */
-  texture = CLUTTER_CAIRO_TEXTURE (clutter_cairo_texture_new (8, 8));
-  cr = clutter_cairo_texture_create (texture);
-
-  pattern = cairo_pattern_create_linear (0, 0, 0, 8);
-  cairo_pattern_add_color_stop_rgba (pattern, 0,
-                                     top->red / 255.,
-                                     top->green / 255.,
-                                     top->blue / 255.,
-                                     top->alpha / 255.);
-  cairo_pattern_add_color_stop_rgba (pattern, 1,
-                                     bottom->red / 255.,
-                                     bottom->green / 255.,
-                                     bottom->blue / 255.,
-                                     bottom->alpha / 255.);
-
-  cairo_set_source (cr, pattern);
-  cairo_paint (cr);
-
-  cairo_pattern_destroy (pattern);
-  cairo_destroy (cr);
-
-  return texture;
-}
-
-/**
- * shell_create_horizontal_gradient:
- * @left: the color on the left
- * @right: the color on the right
- *
- * Creates a horizontal gradient actor.
- *
- * Return value: (transfer none): a #ClutterCairoTexture actor with the
- *               gradient. The texture actor is floating, hence (transfer none).
- */
-ClutterCairoTexture *
-shell_create_horizontal_gradient (ClutterColor *left,
-                                  ClutterColor *right)
-{
-  ClutterCairoTexture *texture;
-  cairo_t *cr;
-  cairo_pattern_t *pattern;
-
-  /* Draw the gradient on an 8x1 pixel texture. Because the gradient is drawn
-   * from the left to the right column, after stretching 1/16 of the
-   * texture width has the left side color and 1/16 has the right side color.
-   * There is no reason to use the 8 pixel height that would be similar to the
-   * reason we are using the 8 pixel width for the vertical gradient, so we
-   * are just using the 1 pixel height instead.
-   */
-  texture = CLUTTER_CAIRO_TEXTURE (clutter_cairo_texture_new (8, 1));
-  cr = clutter_cairo_texture_create (texture);
-
-  pattern = cairo_pattern_create_linear (0, 0, 8, 0);
-  cairo_pattern_add_color_stop_rgba (pattern, 0,
-                                     left->red / 255.,
-                                     left->green / 255.,
-                                     left->blue / 255.,
-                                     left->alpha / 255.);
-  cairo_pattern_add_color_stop_rgba (pattern, 1,
-                                     right->red / 255.,
-                                     right->green / 255.,
-                                     right->blue / 255.,
-                                     right->alpha / 255.);
-
-  cairo_set_source (cr, pattern);
-  cairo_paint (cr);
-
-  cairo_pattern_destroy (pattern);
-  cairo_destroy (cr);
-
-  return texture;
-}
-
 void
 shell_draw_clock (ClutterCairoTexture *texture,
                   int                  hour,
@@ -146,6 +52,70 @@ shell_draw_clock (ClutterCairoTexture *texture,
   cairo_destroy (cr);
 }
 
+/**
+ * shell_fade_app_icon:
+ * @source: Source #ClutterTexture
+ *
+ * Create a new texture by modifying the alpha channel of the
+ * source texture, adding a horizontal gradient fade.
+ *
+ * Returns: (transfer none): A new #ClutterTexture
+ */
+ClutterTexture *
+shell_fade_app_icon (ClutterTexture *source)
+{
+  CoglHandle  texture;
+  guchar     *pixels;
+  gint        width, height, rowstride;
+  gint        fade_start;
+  gint        fade_range;
+  guint       i, j;
+  ClutterTexture *result;
+
+  texture = clutter_texture_get_cogl_texture (source);
+  if (texture == COGL_INVALID_HANDLE)
+    return NULL;
+
+  width  = cogl_texture_get_width  (texture);
+  height = cogl_texture_get_height (texture);
+  rowstride = (width * 4 + 3) & ~3;
+
+  pixels = g_malloc0 (rowstride * height);
+
+  cogl_texture_get_data (texture, COGL_PIXEL_FORMAT_RGBA_8888_PRE,
+                         rowstride, pixels);
+  
+  fade_start = width / 2;
+  fade_range = width - fade_start;
+  for (i = fade_start; i < width; i++)
+    {
+      for (j = 0; j < height; j++)
+        {
+          guchar *pixel = &pixels[j * rowstride + i * 4];
+          float fade = 1.0 - ((float) i - fade_start) / fade_range;
+          pixel[0] = 0.5 + pixel[0] * fade;
+          pixel[1] = 0.5 + pixel[1] * fade;
+          pixel[2] = 0.5 + pixel[2] * fade;
+          pixel[3] = 0.5 + pixel[3] * fade;
+        }   
+    }
+
+  texture = cogl_texture_new_from_data (width,
+                                        height,
+                                        COGL_TEXTURE_NONE,
+                                        COGL_PIXEL_FORMAT_RGBA_8888_PRE,
+                                        COGL_PIXEL_FORMAT_ANY,
+                                        rowstride,
+                                        pixels);
+  g_free (pixels);
+    
+  result = (ClutterTexture*)clutter_texture_new ();
+  clutter_texture_set_cogl_texture (result, texture);
+  cogl_texture_unref (texture);
+  
+  return result;
+}
+
 void
 shell_draw_box_pointer (ClutterCairoTexture   *texture,
                         ShellPointerDirection  direction,
