@@ -1,9 +1,12 @@
 /* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
 
-#include "shell-global.h"
+#include "config.h"
+
+#include "shell-global-private.h"
 #include "shell-wm.h"
 
 #include "display.h"
+#include "util.h"
 #include <clutter/glx/clutter-glx.h>
 #include <clutter/x11/clutter-x11.h>
 #include <gdk/gdkx.h>
@@ -15,13 +18,17 @@
 #include <unistd.h>
 #include <dbus/dbus-glib.h>
 #include <gio/gio.h>
-#include <glib/gi18n-lib.h>
 #include <math.h>
 #include <X11/extensions/Xfixes.h>
+#include <gjs/gjs.h>
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
 
 #define SHELL_DBUS_SERVICE "org.gnome.Shell"
 
 static void grab_notify (GtkWidget *widget, gboolean is_grab, gpointer user_data);
+static void update_root_window_pixmap (ShellGlobal *global);
 
 struct _ShellGlobal {
   GObject parent;
@@ -37,14 +44,18 @@ struct _ShellGlobal {
 
   ShellStageInputMode input_mode;
   XserverRegion input_region;
-  
+
+  GjsContext *js_context;
   MutterPlugin *plugin;
   ShellWM *wm;
+  const char *datadir;
   const char *imagedir;
   const char *configdir;
 
   /* Displays the root window; see shell_global_create_root_pixmap_actor() */
   ClutterActor *root_pixmap;
+
+  gint last_change_screen_width, last_change_screen_height;
 };
 
 enum {
@@ -57,6 +68,7 @@ enum {
   PROP_STAGE,
   PROP_WINDOW_GROUP,
   PROP_WINDOW_MANAGER,
+  PROP_DATADIR,
   PROP_IMAGEDIR,
   PROP_CONFIGDIR,
 };
@@ -64,8 +76,7 @@ enum {
 /* Signals */
 enum
 {
-  PANEL_RUN_DIALOG,
-  PANEL_MAIN_MENU,
+  SCREEN_SIZE_CHANGED,
   LAST_SIGNAL
 };
 
@@ -128,6 +139,9 @@ shell_global_get_property(GObject         *object,
     case PROP_WINDOW_MANAGER:
       g_value_set_object (value, global->wm);
       break;
+    case PROP_DATADIR:
+      g_value_set_string (value, global->datadir);
+      break;
     case PROP_IMAGEDIR:
       g_value_set_string (value, global->imagedir);
       break;
@@ -149,6 +163,7 @@ shell_global_init (ShellGlobal *global)
 
   if (!datadir)
     datadir = GNOME_SHELL_DATADIR;
+  global->datadir = datadir;
 
   /* We make sure imagedir ends with a '/', since the JS won't have
    * access to g_build_filename() and so will end up just
@@ -176,6 +191,9 @@ shell_global_init (ShellGlobal *global)
   global->root_pixmap = NULL;
 
   global->input_mode = SHELL_STAGE_INPUT_MODE_NORMAL;
+
+  global->last_change_screen_width = 0;
+  global->last_change_screen_height = 0;
 }
 
 static void
@@ -186,23 +204,14 @@ shell_global_class_init (ShellGlobalClass *klass)
   gobject_class->get_property = shell_global_get_property;
   gobject_class->set_property = shell_global_set_property;
 
-  shell_global_signals[PANEL_RUN_DIALOG] =
-    g_signal_new ("panel-run-dialog",
+  shell_global_signals[SCREEN_SIZE_CHANGED] =
+    g_signal_new ("screen-size-changed",
 		  G_TYPE_FROM_CLASS (klass),
 		  G_SIGNAL_RUN_LAST,
-		  G_STRUCT_OFFSET (ShellGlobalClass, panel_run_dialog),
+		  G_STRUCT_OFFSET (ShellGlobalClass, screen_size_changed),
 		  NULL, NULL,
-		  g_cclosure_marshal_VOID__INT,
-		  G_TYPE_NONE, 1, G_TYPE_INT);
-
-  shell_global_signals[PANEL_MAIN_MENU] =
-    g_signal_new ("panel-main-menu",
-		  G_TYPE_FROM_CLASS (klass),
-		  G_SIGNAL_RUN_LAST,
-		  G_STRUCT_OFFSET (ShellGlobalClass, panel_main_menu),
-		  NULL, NULL,
-		  g_cclosure_marshal_VOID__INT,
-		  G_TYPE_NONE, 1, G_TYPE_INT);
+		  g_cclosure_marshal_VOID__VOID,
+		  G_TYPE_NONE, 0);
 
   g_object_class_install_property (gobject_class,
                                    PROP_OVERLAY_GROUP,
@@ -254,6 +263,13 @@ shell_global_class_init (ShellGlobalClass *klass)
                                                         "Window management interface",
                                                         SHELL_TYPE_WM,
                                                         G_PARAM_READABLE));
+  g_object_class_install_property (gobject_class,
+                                   PROP_DATADIR,
+                                   g_param_spec_string ("datadir",
+                                                        "Data directory",
+                                                        "Directory containing gnome-shell data files",
+                                                        NULL,
+                                                        G_PARAM_READABLE));
   g_object_class_install_property (gobject_class,
                                    PROP_IMAGEDIR,
                                    g_param_spec_string ("imagedir",
@@ -421,6 +437,73 @@ shell_global_get_windows (ShellGlobal *global)
   return mutter_plugin_get_windows (global->plugin);
 }
 
+static gboolean
+on_screen_size_changed_cb (gpointer data)
+{
+  ShellGlobal *global = SHELL_GLOBAL (data);
+
+  int width, height;
+
+  mutter_plugin_query_screen_size (global->plugin, &width, &height);
+
+  if (global->last_change_screen_width != width || global->last_change_screen_height != height)
+    {
+      g_signal_emit (G_OBJECT (global), shell_global_signals[SCREEN_SIZE_CHANGED], 0);
+      global->last_change_screen_width = width;
+      global->last_change_screen_height = height;
+
+      /* update size of background actor to fix tiled backgrounds */
+      clutter_actor_set_size (CLUTTER_ACTOR (global->root_pixmap),
+                              width, height);
+    }
+
+  return FALSE;
+}
+
+static void
+global_stage_notify_width (GObject    *gobject,
+                           GParamSpec *pspec,
+                           gpointer    data)
+{
+  ShellGlobal *global = SHELL_GLOBAL (data);
+
+  g_object_notify (G_OBJECT (global), "screen-width");
+
+  meta_later_add (META_LATER_BEFORE_REDRAW,
+                  on_screen_size_changed_cb,
+                  global,
+                  NULL);
+}
+
+static void
+global_stage_notify_height (GObject    *gobject,
+                            GParamSpec *pspec,
+                            gpointer    data)
+{
+  ShellGlobal *global = SHELL_GLOBAL (data);
+
+  g_object_notify (G_OBJECT (global), "screen-height");
+
+  meta_later_add (META_LATER_BEFORE_REDRAW,
+                  on_screen_size_changed_cb,
+                  global,
+                  NULL);
+}
+
+static void
+global_plugin_notify_screen (GObject    *gobject,
+                             GParamSpec *pspec,
+                             gpointer    data)
+{
+  ShellGlobal *global = SHELL_GLOBAL (data);
+  ClutterActor *stage = mutter_plugin_get_stage (MUTTER_PLUGIN (gobject));
+
+  g_signal_connect (stage, "notify::width",
+                    G_CALLBACK (global_stage_notify_width), global);
+  g_signal_connect (stage, "notify::height",
+                    G_CALLBACK (global_stage_notify_height), global);
+}
+
 void
 _shell_global_set_plugin (ShellGlobal  *global,
                           MutterPlugin *plugin)
@@ -430,6 +513,21 @@ _shell_global_set_plugin (ShellGlobal  *global,
 
   global->plugin = plugin;
   global->wm = shell_wm_new (plugin);
+
+  /* At this point screen is NULL, so we can't yet do signal connections
+   * to the width and height; we wait until the screen property is set
+   * to do that. Note that this is a one time thing - screen will never
+   * change once first set.
+   */
+  g_signal_connect (plugin, "notify::screen",
+                    G_CALLBACK (global_plugin_notify_screen), global);
+}
+
+void
+_shell_global_set_gjs_context (ShellGlobal *global,
+                               GjsContext  *context)
+{
+  global->js_context = context;
 }
 
 /**
@@ -495,6 +593,70 @@ shell_global_display_is_grabbed (ShellGlobal *global)
   return meta_display_get_grab_op (display) != META_GRAB_OP_NONE;
 }
 
+/* Defining this here for now, see
+ * https://bugzilla.gnome.org/show_bug.cgi?id=604075
+ * for upstreaming status.
+ */
+JSContext * gjs_context_get_context (GjsContext *context);
+
+/**
+ * shell_global_add_extension_importer:
+ * @target_object_script: JavaScript code evaluating to a target object
+ * @target_property: Name of property to use for importer
+ * @directory: Source directory:
+ * @error: A #GError
+ *
+ * This function sets a property named @target_property on the object
+ * resulting from the evaluation of @target_object_script code, which
+ * acts as a GJS importer for directory @directory.
+ *
+ * Returns: %TRUE on success
+ */
+gboolean
+shell_global_add_extension_importer (ShellGlobal *global,
+                                     const char  *target_object_script,
+                                     const char  *target_property,
+                                     const char  *directory,
+                                     GError     **error)
+{
+  jsval target_object;
+  JSObject *importer;
+  JSContext *context = gjs_context_get_context (global->js_context);
+  char *search_path[2] = { 0, 0 };
+
+  // This is a bit of a hack; ideally we'd be able to pass our target
+  // object directly into this function, but introspection doesn't
+  // support that at the moment.  Instead evaluate a string to get it.
+  if (!JS_EvaluateScript(context,
+                         JS_GetGlobalObject(context),
+                         target_object_script,
+                         strlen (target_object_script),
+                         "<target_object_script>",
+                         0,
+                         &target_object))
+    {
+      char *message;
+      gjs_log_exception(context,
+                        &message);
+      g_set_error(error,
+                  G_IO_ERROR,
+                  G_IO_ERROR_FAILED,
+                  "%s", message ? message : "(unknown)");
+      g_free(message);
+      return FALSE;
+    }
+
+  if (!JSVAL_IS_OBJECT (target_object))
+    {
+      g_error ("shell_global_add_extension_importer: invalid target object");
+      return FALSE;
+    }
+
+  search_path[0] = (char*)directory;
+  importer = gjs_define_importer (context, JSVAL_TO_OBJECT (target_object), target_property, (const char **)search_path, FALSE);
+  return TRUE;
+}
+
 /* Code to close all file descriptors before we exec; copied from gspawn.c in GLib.
  *
  * Authors: Padraig O'Briain, Matthias Clasen, Lennart Poettering
@@ -627,7 +789,22 @@ shell_global_reexec_self (ShellGlobal *global)
   g_ptr_array_free (arr, TRUE);
 }
 
-void 
+/**
+ * shell_global_gc:
+ * @global: A #ShellGlobal
+ *
+ * Start a garbage collection process.  For more information, see
+ * https://developer.mozilla.org/En/JS_GC
+ */
+void
+shell_global_gc (ShellGlobal *global)
+{
+  JSContext *context = gjs_context_get_context (global->js_context);
+
+  JS_GC (context);
+}
+
+void
 shell_global_grab_dbus_service (ShellGlobal *global)
 {
   GError *error = NULL;
@@ -660,7 +837,7 @@ shell_global_grab_dbus_service (ShellGlobal *global)
    * unless a special environment variable is passed.  The environment
    * variable is used by the gnome-shell (no --replace) launcher in
    * Xephyr */
-  if (!g_getenv ("GNOME_SHELL_NO_REPLACE_PANEL"))
+  if (!g_getenv ("GNOME_SHELL_NO_REPLACE"))
     {
       if (!dbus_g_proxy_call (bus, "RequestName", &error, G_TYPE_STRING,
                               "org.gnome.Panel", G_TYPE_UINT,
@@ -744,22 +921,6 @@ root_window_filter (GdkXEvent *native, GdkEvent *event, gpointer data)
   return GDK_FILTER_CONTINUE;
 }
 
-/* Workaround for a clutter bug where if ClutterGLXTexturePixmap
- * is painted without the pixmap being set, a crash will occur inside
- * Cogl.
- *
- * http://bugzilla.openedhand.com/show_bug.cgi?id=1644
- */
-static void
-root_pixmap_paint (ClutterActor *actor, gpointer data)
-{
-  Pixmap pixmap;
-
-  g_object_get (G_OBJECT (actor), "pixmap", &pixmap, NULL);
-  if (!pixmap)
-    g_signal_stop_emission_by_name (actor, "paint");
-}
-
 /*
  * Called when the root window pixmap actor is destroyed.
  */
@@ -799,16 +960,24 @@ shell_global_format_time_relative_pretty (ShellGlobal *global,
     *text = g_strdup (_("Less than a minute ago"));
     *next_update = MINUTE - delta;
    } else if (delta < HOUR) {
-     *text = g_strdup_printf (ngettext ("%d minute ago", "%d minutes ago", delta / MINUTE), delta / MINUTE);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d minute ago", "%d minutes ago",
+                                         delta / MINUTE), delta / MINUTE);
      *next_update = MINUTE - (delta % MINUTE);
    } else if (delta < DAY) {
-     *text = g_strdup_printf (ngettext ("%d hour ago", "%d hours ago", delta / HOUR), delta / HOUR);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d hour ago", "%d hours ago",
+                                         delta / HOUR), delta / HOUR);
      *next_update = HOUR - (delta % HOUR);
    } else if (delta < WEEK) {
-     *text = g_strdup_printf (ngettext ("%d day ago", "%d days ago", delta / DAY), delta / DAY);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d day ago", "%d days ago",
+                                         delta / DAY), delta / DAY);
      *next_update = DAY - (delta % DAY);
    } else {
-     *text = g_strdup_printf (ngettext ("%d week ago", "%d weeks ago", delta / WEEK), delta / WEEK);
+     *text = g_strdup_printf (dngettext (GETTEXT_PACKAGE,
+                                         "%d week ago", "%d weeks ago",
+                                         delta / WEEK), delta / WEEK);
      *next_update = WEEK - (delta % WEEK);
    }
 }
@@ -827,6 +996,7 @@ shell_global_create_root_pixmap_actor (ShellGlobal *global)
 {
   GdkWindow *window;
   ClutterActor *stage;
+  ClutterColor stage_color;
 
   /* The actor created is actually a ClutterClone of global->root_pixmap. */
 
@@ -834,20 +1004,32 @@ shell_global_create_root_pixmap_actor (ShellGlobal *global)
     {
       global->root_pixmap = clutter_glx_texture_pixmap_new ();
 
+      clutter_texture_set_repeat (CLUTTER_TEXTURE (global->root_pixmap),
+                                  TRUE, TRUE);
+
       /* The low and medium quality filters give nearest-neighbor resizing. */
       clutter_texture_set_filter_quality (CLUTTER_TEXTURE (global->root_pixmap),
                                           CLUTTER_TEXTURE_QUALITY_HIGH);
 
+      /* Initialize to the stage color, since that's what will be seen
+       * in the main view if there's no actual background window.
+       */
+      stage = mutter_plugin_get_stage (global->plugin);
+      clutter_stage_get_color (CLUTTER_STAGE (stage), &stage_color);
+      clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (global->root_pixmap),
+                                         /* ClutterColor has the same layout
+                                          * as one pixel of RGB(A) data.
+                                          */
+                                         (const guchar *)&stage_color, FALSE,
+                                         /* w, h, rowstride, bpp, flags */
+                                         1, 1, 3, 3, 0, NULL);
+
       /* We can only clone an actor within a stage, so we hide the source
        * texture then add it to the stage */
       clutter_actor_hide (global->root_pixmap);
-      stage = mutter_plugin_get_stage (global->plugin);
       clutter_container_add_actor (CLUTTER_CONTAINER (stage),
                                    global->root_pixmap);
 
-      g_signal_connect (global->root_pixmap, "paint",
-                        G_CALLBACK (root_pixmap_paint), NULL);
-
       /* This really should never happen; but just in case... */
       g_signal_connect (global->root_pixmap, "destroy",
                         G_CALLBACK (root_pixmap_destroy), global);
@@ -869,3 +1051,231 @@ shell_global_create_root_pixmap_actor (ShellGlobal *global)
 
   return clutter_clone_new (global->root_pixmap);
 }
+
+/**
+ * shell_global_get_monitors:
+ * @global: the #ShellGlobal
+ *
+ * Gets a list of the bounding boxes of the active screen's monitors.
+ *
+ * Return value: (transfer full) (element-type GdkRectangle): a list
+ * of monitor bounding boxes.
+ */
+GSList *
+shell_global_get_monitors (ShellGlobal *global)
+{
+  MetaScreen *screen = shell_global_get_screen (global);
+  GSList *monitors = NULL;
+  MetaRectangle rect;
+  int i;
+
+  g_assert (sizeof (MetaRectangle) == sizeof (GdkRectangle) &&
+            G_STRUCT_OFFSET (MetaRectangle, x) == G_STRUCT_OFFSET (GdkRectangle, x) &&
+            G_STRUCT_OFFSET (MetaRectangle, y) == G_STRUCT_OFFSET (GdkRectangle, y) &&
+            G_STRUCT_OFFSET (MetaRectangle, width) == G_STRUCT_OFFSET (GdkRectangle, width) &&
+            G_STRUCT_OFFSET (MetaRectangle, height) == G_STRUCT_OFFSET (GdkRectangle, height));
+
+  for (i = meta_screen_get_n_monitors (screen) - 1; i >= 0; i--)
+    {
+      meta_screen_get_monitor_geometry (screen, i, &rect);
+      monitors = g_slist_prepend (monitors,
+                                  g_boxed_copy (GDK_TYPE_RECTANGLE, &rect));
+    }
+  return monitors;
+}
+
+/**
+ * shell_global_get_primary_monitor:
+ * @global: the #ShellGlobal
+ *
+ * Gets the bounding box of the primary monitor (the one that the
+ * panel is on).
+ *
+ * Return value: the bounding box of the primary monitor
+ */
+GdkRectangle *
+shell_global_get_primary_monitor (ShellGlobal  *global)
+{
+  MetaScreen *screen = shell_global_get_screen (global);
+  MetaRectangle rect;
+
+  g_assert (sizeof (MetaRectangle) == sizeof (GdkRectangle) &&
+            G_STRUCT_OFFSET (MetaRectangle, x) == G_STRUCT_OFFSET (GdkRectangle, x) &&
+            G_STRUCT_OFFSET (MetaRectangle, y) == G_STRUCT_OFFSET (GdkRectangle, y) &&
+            G_STRUCT_OFFSET (MetaRectangle, width) == G_STRUCT_OFFSET (GdkRectangle, width) &&
+            G_STRUCT_OFFSET (MetaRectangle, height) == G_STRUCT_OFFSET (GdkRectangle, height));
+
+  meta_screen_get_monitor_geometry (screen, 0, &rect);
+  return g_boxed_copy (GDK_TYPE_RECTANGLE, &rect);
+}
+
+/**
+ * shell_global_get_focus_monitor:
+ * @global: the #ShellGlobal
+ *
+ * Gets the bounding box of the monitor containing the window that
+ * currently contains the keyboard focus.
+ *
+ * Return value: the bounding box of the focus monitor
+ */
+GdkRectangle *
+shell_global_get_focus_monitor (ShellGlobal  *global)
+{
+  MetaScreen *screen = shell_global_get_screen (global);
+  MetaDisplay *display = meta_screen_get_display (screen);
+  MetaWindow *focus = meta_display_get_focus_window (display);
+  MetaRectangle rect, wrect;
+  int nmonitors, i;
+
+  if (focus)
+    {
+      meta_window_get_outer_rect (focus, &wrect);
+      nmonitors = meta_screen_get_n_monitors (screen);
+
+      /* Find the monitor that the top-left corner of @focus is on. */
+      for (i = 0; i < nmonitors; i++)
+        {
+          meta_screen_get_monitor_geometry (screen, i, &rect);
+
+          if (rect.x < wrect.x && rect.y < wrect.y &&
+              rect.x + rect.width > wrect.x &&
+              rect.y + rect.height > wrect.y)
+            return g_boxed_copy (GDK_TYPE_RECTANGLE, &rect);
+        }
+    }
+
+  meta_screen_get_monitor_geometry (screen, 0, &rect);
+  return g_boxed_copy (GDK_TYPE_RECTANGLE, &rect);
+}
+
+/**
+ * shell_global_get_modifier_keys:
+ * @global: the #ShellGlobal
+ *
+ * Gets the current set of modifier keys that are pressed down;
+ * this is a wrapper around gdk_display_get_pointer() that strips
+ * out any un-declared modifier flags, to make gjs happy; see
+ * https://bugzilla.gnome.org/show_bug.cgi?id=597292.
+ *
+ * Return value: the current modifiers
+ */
+GdkModifierType
+shell_global_get_modifier_keys (ShellGlobal *global)
+{
+  GdkModifierType mods;
+
+  gdk_display_get_pointer (gdk_display_get_default (), NULL, NULL, NULL, &mods);
+  return mods & GDK_MODIFIER_MASK;
+}
+
+/**
+ * shell_get_event_state:
+ * @event: a #ClutterEvent
+ *
+ * Gets the current state of the event (the set of modifier keys that
+ * are pressed down). Thhis is a wrapper around
+ * clutter_event_get_state() that strips out any un-declared modifier
+ * flags, to make gjs happy; see
+ * https://bugzilla.gnome.org/show_bug.cgi?id=597292.
+ *
+ * Return value: the state from the event
+ */
+ClutterModifierType
+shell_get_event_state (ClutterEvent *event)
+{
+  ClutterModifierType state = clutter_event_get_state (event);
+  return state & CLUTTER_MODIFIER_MASK;
+}
+
+static void
+shell_popup_menu_position_func (GtkMenu   *menu,
+                                int       *x,
+                                int       *y,
+                                gboolean  *push_in,
+                                gpointer   user_data)
+{
+  *x = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (menu), "shell-menu-x"));
+  *y = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (menu), "shell-menu-y"));
+}
+
+/**
+ * shell_popup_menu:
+ * @menu: a #GtkMenu
+ * @button: mouse button that triggered the menu
+ * @time: timestamp of event that triggered the menu
+ * @menu_x: x coordinate to display the menu at
+ * @menu_y: y coordinate to display the menu at
+ *
+ * Wraps gtk_menu_popup(), but using @menu_x, @menu_y for the location
+ * rather than needing a callback.
+ **/
+void
+shell_popup_menu (GtkMenu *menu, int button, guint32 time,
+                  int menu_x, int menu_y)
+{
+  g_object_set_data (G_OBJECT (menu), "shell-menu-x", GINT_TO_POINTER (menu_x));
+  g_object_set_data (G_OBJECT (menu), "shell-menu-y", GINT_TO_POINTER (menu_y));
+
+  gtk_menu_popup (menu, NULL, NULL, shell_popup_menu_position_func, NULL,
+                  button, time);
+}
+
+/**
+ * shell_global_get_current_time:
+ * @global: A #ShellGlobal
+ *
+ * Returns: the current X server time from the current Clutter, Gdk, or X
+ * event. If called from outside an event handler, this may return
+ * %Clutter.CURRENT_TIME (aka 0), or it may return a slightly
+ * out-of-date timestamp.
+ */
+guint32
+shell_global_get_current_time (ShellGlobal *global)
+{
+  guint32 time;
+  MetaDisplay *display;
+
+  /* meta_display_get_current_time() will return the correct time
+     when handling an X or Gdk event, but will return CurrentTime
+     from some Clutter event callbacks.
+
+     clutter_get_current_event_time() will return the correct time
+     from a Clutter event callback, but may return an out-of-date
+     timestamp if called at other times.
+
+     So we try meta_display_get_current_time() first, since we
+     can recognize a "wrong" answer from that, and then fall back
+     to clutter_get_current_event_time().
+   */
+
+  display = meta_screen_get_display (shell_global_get_screen (global));
+  time = meta_display_get_current_time (display);
+  if (time != CLUTTER_CURRENT_TIME)
+      return time;
+
+  return clutter_get_current_event_time ();
+}
+
+/**
+ * shell_global_get_app_launch_context:
+ * @global: A #ShellGlobal
+ *
+ * Create a #GAppLaunchContext set up with the correct timestamp, and
+ * targeted to activate on the current workspace.
+ *
+ * Return value: A new #GAppLaunchContext
+ */
+GAppLaunchContext *
+shell_global_create_app_launch_context (ShellGlobal *global)
+{
+  GdkAppLaunchContext *context;
+
+  context = gdk_app_launch_context_new ();
+  gdk_app_launch_context_set_timestamp (context, shell_global_get_current_time (global));
+
+  // Make sure that the app is opened on the current workspace even if
+  // the user switches before it starts
+  gdk_app_launch_context_set_desktop (context, meta_screen_get_active_workspace_index (shell_global_get_screen (global)));
+
+  return (GAppLaunchContext *)context;
+}
