@@ -1,3 +1,4 @@
+/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
 #ifndef __SHELL_GLOBAL_H__
 #define __SHELL_GLOBAL_H__
 
@@ -25,10 +26,7 @@ struct _ShellGlobalClass
 {
   GObjectClass parent_class;
 
-  void (*panel_run_dialog) (ShellGlobal *global,
-			    int      timestamp);
-  void (*panel_main_menu) (ShellGlobal *global,
-			   int      timestamp);
+  void (*screen_size_changed) (ShellGlobal *global);
 };
 
 GType            shell_global_get_type            (void) G_GNUC_CONST;
@@ -42,6 +40,12 @@ ShellGlobal *shell_global_get (void);
 
 MetaScreen *shell_global_get_screen (ShellGlobal  *global);
 
+gboolean shell_global_add_extension_importer (ShellGlobal *global,
+                                              const char  *target_object_script,
+                                              const char  *target_property,
+                                              const char  *directory,
+                                              GError     **error);
+
 void shell_global_grab_dbus_service (ShellGlobal *global);
 
 typedef enum {
@@ -57,9 +61,6 @@ void shell_global_set_stage_input_region (ShellGlobal         *global,
 
 GList *shell_global_get_windows (ShellGlobal *global);
 
-void _shell_global_set_plugin (ShellGlobal  *global,
-			       MutterPlugin *plugin);
-
 gboolean shell_global_begin_modal (ShellGlobal *global,
 				   guint32      timestamp);
 void     shell_global_end_modal   (ShellGlobal *global,
@@ -69,10 +70,27 @@ gboolean shell_global_display_is_grabbed (ShellGlobal *global);
 
 void shell_global_reexec_self (ShellGlobal *global);
 
+void shell_global_gc (ShellGlobal *global);
+
 void shell_global_format_time_relative_pretty (ShellGlobal *global, guint delta, char **text, guint *update_time);
 
 ClutterActor *shell_global_create_root_pixmap_actor (ShellGlobal *global);
 
+GSList       *shell_global_get_monitors        (ShellGlobal  *global);
+GdkRectangle *shell_global_get_primary_monitor (ShellGlobal  *global);
+GdkRectangle *shell_global_get_focus_monitor   (ShellGlobal  *global);
+
+GdkModifierType shell_global_get_modifier_keys (ShellGlobal *global);
+
+ClutterModifierType shell_get_event_state (ClutterEvent *event);
+
+void shell_popup_menu (GtkMenu *menu, int button, guint32 time,
+                       int menu_x, int menu_y);
+
+guint32 shell_global_get_current_time (ShellGlobal *global);
+
+GAppLaunchContext *shell_global_create_app_launch_context (ShellGlobal *global);
+
 G_END_DECLS
 
 #endif /* __SHELL_GLOBAL_H__ */
