@@ -26,10 +26,17 @@ GType st_corner_get_type (void) G_GNUC_CONST;
 GType st_text_decoration_get_type (void) G_GNUC_CONST;
 #define ST_TYPE_TEXT_DECORATION (st_text_decoration_get_type())
 
+GType st_gradient_type_get_type (void) G_GNUC_CONST;
+#define ST_TYPE_GRADIENT_TYPE (st_gradient_type_get_type())
+
 /* enumerations from "st/st-types.h" */
 GType st_align_get_type (void) G_GNUC_CONST;
 #define ST_TYPE_ALIGN (st_align_get_type())
 
+/* enumerations from "st/st-widget.h" */
+GType st_text_direction_get_type (void) G_GNUC_CONST;
+#define ST_TYPE_TEXT_DIRECTION (st_text_direction_get_type())
+
 G_END_DECLS
 
 #endif /* !__ST_ENUM_TYPES_H__ */
